@@ -42,6 +42,22 @@ except ImportError:
 
 logger = logging.getLogger(__name__)
 
+PIL_IMAGE_FORMATS = Pil.registered_extensions()
+
+
+def _pprint_pil_formats():
+    res = ''
+    row = []
+    for i in PIL_IMAGE_FORMATS:
+        if len(row) > 12:
+            res = res + ", ".join(row) + ",\n"
+            row = []
+        row.append(i)
+    return res + ", ".join(row)
+
+
+PIL_IMAGE_FORMATS_STR = _pprint_pil_formats()
+
 
 def ensure_dir(filename):
     """Checks if the dir of f exists, otherwise create it.
@@ -58,30 +74,17 @@ class UnknownImageFormat(Exception):
 
 
 def check_image_format(fformat):
-    """Check that *fformat* is valid
+    """Check that *fformat* is valid.
+
+    Valid formats are listed in https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html
     """
-    cases = {"jpg": "jpeg",
-             "jpeg": "jpeg",
-             "tif": "tiff",
-             "tiff": "tif",
-             "pgm": "ppm",
-             "pbm": "ppm",
-             "ppm": "ppm",
-             "bmp": "bmp",
-             "dib": "bmp",
-             "gif": "gif",
-             "im": "im",
-             "pcx": "pcx",
-             "png": "png",
-             "xbm": "xbm",
-             "xpm": "xpm",
-             'jp2': 'jp2',
-             }
     fformat = fformat.lower()
     try:
-        fformat = cases[fformat]
+        fformat = PIL_IMAGE_FORMATS["." + fformat]
     except KeyError:
-        raise UnknownImageFormat("Unknown image format '%s'." % fformat)
+        raise UnknownImageFormat(
+            "Unknown image format '%s'.  Supported formats for 'simple_image' writer are:\n%s" %
+            (fformat, PIL_IMAGE_FORMATS_STR))
     return fformat
 
 
@@ -178,7 +181,7 @@ class Image(object):
                         color_min = color_range[i][0]
                         color_max = color_range[i][1]
                         # Add data to image object as a channel
-                        #self._add_channel(chn, color_min, color_max)
+                        # self._add_channel(chn, color_min, color_max)
                     else:
                         color_min = 0.0
                         color_max = 1.0
@@ -366,9 +369,12 @@ class Image(object):
 
     def pil_save(self, filename, compression=6, fformat=None,
                  thumbnail_name=None, thumbnail_size=None):
-        """Save the image to the given *filename* using PIL. For now, the
-        compression level [0-9] is ignored, due to PIL's lack of support. See
-        also :meth:`save`.
+        """Save the image to the given *filename* using PIL.
+
+        For now, the compression level [0-9] is ignored, due to PIL's lack of support.
+        See also :meth:`save`.
+
+        Supported image formats are listed in https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html
         """
         # PIL does not support compression option.
         del compression
@@ -384,12 +390,12 @@ class Image(object):
 
         params = {}
 
-        if fformat == 'png':
+        if fformat == 'PNG':
             # Take care of GeoImage.tags (if any).
             params['pnginfo'] = self._pngmeta()
 
         # JPEG images does not support transparency
-        if fformat == 'jpeg' and not self.fill_value:
+        if fformat == 'JPEG' and not self.fill_value:
             self.fill_value = [0, 0, 0, 0]
             logger.debug("No fill_value provided, setting it to 0.")
 
