@@ -35,6 +35,7 @@ chunks can be saved in parallel.
 import logging
 import os
 import threading
+import warnings
 from contextlib import suppress
 
 import dask
@@ -414,19 +415,20 @@ class XRImage(object):
         kwformat = format_kwargs.pop('format', None)
         fformat = fformat or kwformat or os.path.splitext(filename)[1][1:]
         if fformat in ('tif', 'tiff', 'jp2') and rasterio:
+
             return self.rio_save(filename, fformat=fformat,
                                  fill_value=fill_value, compute=compute,
                                  keep_palette=keep_palette, cmap=cmap,
                                  **format_kwargs)
-        else:
-            return self.pil_save(filename, fformat, fill_value,
-                                 compute=compute, **format_kwargs)
+        return self.pil_save(filename, fformat, fill_value,
+                             compute=compute, **format_kwargs)
 
     def rio_save(self, filename, fformat=None, fill_value=None,
                  dtype=np.uint8, compute=True, tags=None,
                  keep_palette=False, cmap=None, overviews=None,
                  overviews_minsize=256, overviews_resampling=None,
                  include_scale_offset_tags=False,
+                 scale_offset_tags=None,
                  **format_kwargs):
         """Save the image using rasterio.
 
@@ -463,9 +465,10 @@ class XRImage(object):
                 provided. Common values include `nearest` (default),
                 `bilinear`, `average`, and many others. See the rasterio
                 documentation for more information.
-            include_scale_offset_tags (bool): Whether or not (default) to
-                include a ``scale`` and an ``offset`` tag in the data that would
-                help retrieving original data values from pixel values.
+            scale_offset_tags (Tuple[str, str] or None)
+                If set to a ``(str, str)`` tuple, scale and offset will be
+                stored in GDALMetaData tags.  Those can then be used to
+                retrieve the original data values from pixel values.
 
         Returns:
             The delayed or computed result of the saving.
@@ -478,12 +481,18 @@ class XRImage(object):
                    'tiff': 'GTiff',
                    'jp2': 'JP2OpenJPEG'}
         driver = drivers.get(fformat, fformat)
+        if include_scale_offset_tags:
+            warnings.warn(
+                "include_scale_offset_tags is deprecated, please use "
+                "scale_offset_tags to indicate tag labels",
+                DeprecationWarning)
+            scale_offset_tags = scale_offset_tags or ("scale", "offset")
 
         if tags is None:
             tags = {}
 
         data, mode = self.finalize(fill_value, dtype=dtype,
-                                   keep_palette=keep_palette, cmap=cmap)
+                                   keep_palette=keep_palette)
         data = data.transpose('bands', 'y', 'x')
 
         crs = None
@@ -536,9 +545,10 @@ class XRImage(object):
         elif driver == 'JPEG' and 'A' in mode:
             raise ValueError('JPEG does not support alpha')
 
-        if include_scale_offset_tags:
+        if scale_offset_tags:
+            scale_label, offset_label = scale_offset_tags
             scale, offset = self.get_scaling_from_history(data.attrs.get('enhancement_history', []))
-            tags['scale'], tags['offset'] = invert_scale_offset(scale, offset)
+            tags[scale_label], tags[offset_label] = invert_scale_offset(scale, offset)
 
         # FIXME add metadata
         r_file = RIOFile(filename, 'w', driver=driver,
@@ -564,14 +574,7 @@ class XRImage(object):
             except AttributeError:
                 raise ValueError("Colormap is not formatted correctly")
 
-        da_tags = []
-        for key, val in list(tags.items()):
-            try:
-                if isinstance(val.data, da.Array):
-                    da_tags.append((val.data, RIOTag(r_file, key)))
-                    tags.pop(key)
-            except AttributeError:
-                continue
+        tags, da_tags = self._split_regular_vs_lazy_tags(tags, r_file)
 
         r_file.rfile.update_tags(**tags)
         r_dataset = RIODataset(r_file, overviews,
@@ -596,6 +599,21 @@ class XRImage(object):
         # closing the file
         return to_store
 
+    @staticmethod
+    def _split_regular_vs_lazy_tags(tags, r_file):
+        """Split tags into regular vs lazy (dask) tags."""
+        da_tags = []
+        for key, val in list(tags.items()):
+            try:
+                if isinstance(val.data, da.Array):
+                    da_tags.append((val.data, RIOTag(r_file, key)))
+                    tags.pop(key)
+                else:
+                    tags[key] = val.item()
+            except AttributeError:
+                continue
+        return tags, da_tags
+
     def pil_save(self, filename, fformat=None, fill_value=None,
                  compute=True, **format_kwargs):
         """Save the image to the given *filename* using PIL.
@@ -746,7 +764,9 @@ class XRImage(object):
         null_mask = null_mask.any(dim='bands')
         null_mask = null_mask.expand_dims('bands')
         null_mask['bands'] = ['A']
-        # match data dtype
+        # changes to null_mask attrs should not effect the original attrs
+        # XRImage never uses them either
+        null_mask.attrs = {}
         return null_mask
 
     def _add_alpha(self, data, alpha=None):
@@ -773,7 +793,27 @@ class XRImage(object):
         data.attrs = attrs
         return data
 
-    def _scale_to_dtype(self, data, dtype):
+    def _get_dtype_scale_offset(self, dtype, fill_value):
+        dinfo = np.iinfo(dtype)
+        scale = dinfo.max - dinfo.min
+        offset = dinfo.min
+        if fill_value is not None:
+            if fill_value == dinfo.min:
+                # leave the lowest value for fill value only
+                offset = offset + 1
+                scale = scale - 1
+            elif fill_value == dinfo.max:
+                # leave the top value for fill value only
+                scale = scale - 1
+            else:
+                warnings.warn(
+                    "Specified fill value will overlap with valid "
+                    "data. To avoid this warning specify a fill_value "
+                    "that is the minimum or maximum for the data type "
+                    "being saved to.")
+        return scale, offset
+
+    def _scale_to_dtype(self, data, dtype, fill_value=None):
         """Scale provided data to dtype range assuming a 0-1 range.
 
         Float input data is assumed to be normalized to a 0 to 1 range.
@@ -789,9 +829,8 @@ class XRImage(object):
                 data = data.clip(np.iinfo(dtype).min, np.iinfo(dtype).max)
             else:
                 # scale float data (assumed to be 0 to 1) to full integer space
-                dinfo = np.iinfo(dtype)
-                scale = dinfo.max - dinfo.min
-                offset = dinfo.min
+                # leave room for fill value if needed
+                scale, offset = self._get_dtype_scale_offset(dtype, fill_value)
                 data = data.clip(0, 1) * scale + offset
                 attrs.setdefault('enhancement_history', list()).append({'scale': scale, 'offset': offset})
             data = data.round()
@@ -899,21 +938,54 @@ class XRImage(object):
             new_img.palette = self.palette
         return new_img
 
-    def _finalize(self, fill_value=None, dtype=np.uint8, keep_palette=False, cmap=None):
-        """Wrap around 'finalize' method for backwards compatibility."""
-        import warnings
-        warnings.warn("'_finalize' is deprecated, use 'finalize' instead.",
-                      DeprecationWarning)
-        return self.finalize(fill_value, dtype, keep_palette, cmap)
-
     def final_mode(self, fill_value=None):
         """Get the mode of the finalized image when provided this fill_value."""
         if fill_value is None and not self.mode.endswith('A'):
             return self.mode + 'A'
-        else:
-            return self.mode
+        return self.mode
 
-    def finalize(self, fill_value=None, dtype=np.uint8, keep_palette=False, cmap=None):
+    def _add_alpha_and_scale(self, data, ifill, dtype):
+        alpha = self._create_alpha(data, fill_value=ifill)
+        data = self._scale_to_dtype(data, dtype)
+        data = data.astype(dtype)
+        data = self._add_alpha(data, alpha=alpha)
+        return data
+
+    def _replace_fill_value(self, data, ifill, fill_value, dtype):
+        # Add fill_value after all other calculations have been done to
+        # make sure it is not scaled for the data type
+        if ifill is not None and fill_value is not None:
+            # cast fill value to output type so we don't change data type
+            fill_value = dtype(fill_value)
+            # integer fields have special fill values
+            data = data.where(data != ifill, dtype(fill_value))
+        elif fill_value is not None:
+            data = data.fillna(dtype(fill_value))
+
+        return data
+
+    def _get_input_fill_value(self, data):
+        # if the data are integers then this fill value will be used to check for invalid values
+        if np.issubdtype(data, np.integer):
+            return data.attrs.get('_FillValue')
+        return None
+
+    def _scale_and_replace_fill_value(self, data, input_fill_value, fill_value, dtype):
+        # scale float data to the proper dtype
+        # this method doesn't cast yet so that we can keep track of NULL values
+        data = self._scale_to_dtype(data, dtype, fill_value)
+        data = self._replace_fill_value(data, input_fill_value, fill_value, dtype)
+        return data
+
+    def _scale_alpha_or_fill_data(self, data, fill_value, dtype):
+        input_fill_value = self._get_input_fill_value(data)
+        needs_alpha = fill_value is None and not self.mode.endswith('A')
+        if needs_alpha:
+            # We don't have a fill value or an alpha, let's add an alpha
+            return self._add_alpha_and_scale(data, input_fill_value, dtype)
+        return self._scale_and_replace_fill_value(data, input_fill_value, fill_value, dtype)
+
+    def finalize(self, fill_value=None, dtype=np.uint8, keep_palette=False):
         """Finalize the image to be written to an output file.
 
         This adds an alpha band or fills data with a fill_value (if
@@ -925,17 +997,44 @@ class XRImage(object):
         determined by a special ``_FillValue`` attribute in the
         ``DataArray`` ``.attrs`` dictionary.
 
+        Args:
+            fill_value (int or float or None): Output value to use to
+                represent invalid or missing pixels. By default this is
+                `None` meaning an Alpha channel will be used to represent
+                the invalid values; transparent for invalid, opaque
+                otherwise. Some output formats do not support alpha channels
+                so a ``fill_value`` must be provided. This is determined by
+                the underlying library doing the writing (pillow or rasterio).
+                If specified, it should be the minimum or maximum of the
+                ``dtype`` (ex. 0 or 255 for uint8). Floating point image data
+                is then scaled to fit the remainder of the data type space.
+                Integer image data will **not** be scaled. For example, a
+                ``dtype`` of ``numpy.uint8`` and a ``fill_value`` of 0 will
+                result in floating-point data being scaled linearly from 1 to 255.
+            dtype (numpy.dtype): Output data type to convert the current image
+                data to. Default is unsigned 8-bit integer
+                (:class:`numpy.uint8`).
+            keep_palette (bool): Whether to convert a paletted image to RGB/A
+                or not. If ``False`` (default) then ``P`` mode images will be
+                converted to ``RGB`` and ``PA`` will be converted to ``RGBA``.
+                If ``True``, images with mode ``P`` or ``PA`` are kept as is
+                and will not be scaled in order for their index values into a
+                palette to be maintained. This flag should always be ``False``
+                for non-paletted images.
+
         """
         if keep_palette and not self.mode.startswith('P'):
             keep_palette = False
 
         if not keep_palette:
+            finalize_kwargs = dict(
+                fill_value=fill_value, dtype=dtype,
+                keep_palette=keep_palette,
+            )
             if self.mode == "P":
-                return self.convert("RGB").finalize(fill_value=fill_value, dtype=dtype,
-                                                    keep_palette=keep_palette, cmap=cmap)
+                return self.convert("RGB").finalize(**finalize_kwargs)
             if self.mode == "PA":
-                return self.convert("RGBA").finalize(fill_value=fill_value, dtype=dtype,
-                                                     keep_palette=keep_palette, cmap=cmap)
+                return self.convert("RGBA").finalize(**finalize_kwargs)
 
         if np.issubdtype(dtype, np.floating) and fill_value is None:
             logger.warning("Image with floats cannot be transparent, so "
@@ -947,34 +1046,10 @@ class XRImage(object):
             final_data.attrs['enhancement_history'] = list(self.data.attrs['enhancement_history'])
         except KeyError:
             pass
-        attrs = final_data.attrs
-        # if the data are integers then this fill value will be used to check for invalid values
         with xr.set_options(keep_attrs=True):
-            ifill = final_data.attrs.get('_FillValue') if np.issubdtype(final_data, np.integer) else None
+            attrs = final_data.attrs
             if not keep_palette:
-                if fill_value is None and not self.mode.endswith('A'):
-                    # We don't have a fill value or an alpha, let's add an alpha
-                    alpha = self._create_alpha(final_data, fill_value=ifill)
-                    final_data = self._scale_to_dtype(final_data, dtype)
-                    attrs = final_data.attrs
-                    final_data = final_data.astype(dtype)
-                    final_data = self._add_alpha(final_data, alpha=alpha)
-                    final_data.attrs = attrs
-                else:
-                    # scale float data to the proper dtype
-                    # this method doesn't cast yet so that we can keep track of NULL values
-                    final_data = self._scale_to_dtype(final_data, dtype)
-                    attrs = final_data.attrs
-                    # Add fill_value after all other calculations have been done to
-                    # make sure it is not scaled for the data type
-                    if ifill is not None and fill_value is not None:
-                        # cast fill value to output type so we don't change data type
-                        fill_value = dtype(fill_value)
-                        # integer fields have special fill values
-                        final_data = final_data.where(final_data != ifill, dtype(fill_value))
-                    elif fill_value is not None:
-                        final_data = final_data.fillna(dtype(fill_value))
-
+                final_data = self._scale_alpha_or_fill_data(final_data, fill_value, dtype)
             final_data = final_data.astype(dtype)
             final_data.attrs = attrs
 
@@ -1394,10 +1469,10 @@ class XRImage(object):
             raise ValueError(
                     "Expected self.mode='RGBA', got {md!s}".format(
                         md=self.mode))
-        elif not isinstance(src, XRImage):
+        if not isinstance(src, XRImage):
             raise TypeError("Expected XRImage, got {tp!s}".format(
                 tp=type(src)))
-        elif src.mode != "RGBA":
+        if src.mode != "RGBA":
             raise ValueError("Expected src.mode='RGBA', got {sm!s}".format(
                 sm=src.mode))
 
